<?php

class Wizard_Form {
  var $fapi_form = NULL, // Original Forms API form pointer.
      $wizard_form = NULL, // Freshly sliced Forms API form.
      $current_page = -1, // Current Wizard Form page number.
      $pages = -1; // Array of Wizard Form pages.

  /**
   * Class Constructor.
   */
  function __construct(&$form) {
    $this->fapi_form = &$form;
  }

  /**
   * Get the current Wizard Form page number;
   * it will either be provided in a hidden form element (which needs to be validated),
   * or we will assume it is the first page.
   *
   * @return Integer
   *   Current page number.
   */
  function get_current_page() {
    // Return cached value, if possible.
    if ($this->current_page > -1) return $this->current_page;

    $this->set_current_page(isset($this->fapi_form['#post']['page'])? (int) $this->fapi_form['#post']['page'] : 0);

    return $this->current_page;
  }

  /**
   * Set the current Wizard Form page number,
   * and validate that it exists.
   *
   * @param Integer $current_page
   *   Current page number.
   * @return Integer
   *   Current page number.
   */
  function set_current_page($current_page) {
    // Be sure the current Wizard Form page falls within range of the total.
    if ($current_page >= 0 && $current_page <= $this->get_pages()) {
      $this->current_page = (int) $current_page;
    }

    return $this->get_current_page();
  }

  /**
   * Get the total number of Wizard Form pages; and/or
   * Count the number of CCK Wizard fields (a.k.a. page-breaks) in a given form,
   * and slice the form elements into separate Wizard Form pages.
   *
   * @return Integer
   *   Total number of Wizard Form pages.
   */
  function get_pages() {
    // Return cached value, if possible.
    if ($this->pages > -1) return $this->pages;

    // Sort FAPI form elements by weight first.
    if (!isset($this->fapi_form['#sorted'])) {
      uasort($this->fapi_form, '_element_sort');
      $this->fapi_form['#sorted'] = TRUE;
    }
    $this->pages = 0;

    foreach (array_keys($this->fapi_form) as $k) {
      // Gather special form elements which should exist in every Wizard Form page,
      // and will not affect form validation when the form is split into different pages.
      // TODO: Should this be recursing into child elements?
      if (element_property($k) || // Include properties of a structured array element.
          (isset($this->fapi_form[$k]['#type']) && in_array($this->fapi_form[$k]['#type'], array('hidden', 'token', 'value'))) // Include token, value, and hidden form elements.
      ) { $this->wizard_form['persistant'][$k] = &$this->fapi_form[$k]; continue; }

      // Separate form into pages by splitting at each CCK Wizard field.
      if (isset($this->fapi_form[$k]['wizard'])) { $this->pages++; continue; }

      // Skip the Preview and Submit buttons; these will be replaced later on.
      if (($k == 'preview' && isset($this->fapi_form[$k]['#type']) && $this->fapi_form[$k]['#type'] = 'button' && isset($this->fapi_form[$k]['#value']) && $this->fapi_form[$k]['#value'] = 'Preview') ||
          ($k == 'submit' && isset($this->fapi_form[$k]['#type']) && $this->fapi_form[$k]['#type'] = 'button' && isset($this->fapi_form[$k]['#value']) && $this->fapi_form[$k]['#value'] = 'Submit')
      ) { continue; }

      // Assign remaining fields to the appropriate Wizard Form page.
      $this->wizard_form[$this->pages][$k] = &$this->fapi_form[$k];
    }

    return $this->pages;
  }

  /**
   * Return an altered version of the original Forms API form,
   * only containing form elements assigned to the current Wizard Form page.
   *
   * @return Array
   *   Altered Forms API form.
   */
  function alter_form() {
    if (($pages = $this->get_pages()) < 1) return; // Need at least two pages before we can alter the original FAPI form.
    $current_page = $this->get_current_page();
    $form = array();

    // Check for signal/authorization to move to another page from wizard_submit().
    if (isset($this->fapi_form['#wizard_move_to'])) {
      $current_page = $this->set_current_page($this->fapi_form['#wizard_move_to']); // Move to another page.
      $form['#wizard_move_to'] = &$this->fapi_form['#wizard_move_to']; // remember to prevent recursion past 2nd level
    }

    // Begin adding only the special form elements (properties of a structured array element);
    // these should be present on every Wizard Form page.
    $form = array_merge($form, &$this->wizard_form['persistant']);

    // Add fields from the current Wizard Form page.
    $form = array_merge($form, &$this->wizard_form[$current_page]);

    // Convert submitted form values from other pages to hidden form elements.
    // when we are instructed to move to a diff. page
    // or when we begin on a page other than the first page and POSTed form values are present
    if ((isset($this->fapi_form['#wizard_move_to']) && isset($this->fapi_form['#post']))
        || ($current_page > 0 && isset($this->fapi_form['#post']))) {
      $form = array_merge($form, $this->hide_submitted($form, $this->fapi_form['#post']));
    }

    // Store the page number in a hidden form element.
    $form['page'] = array('#type' => 'hidden', '#value' => $current_page);

    // Store the total number of pages in a value.
    $form['pages'] = array('#type' => 'value', '#value' => $pages);

    // Setting #multistep to true activates some special form
    // handling code for Drupal. First, the FormAPI makes sure
    // that the results of the form's submission are passed back
    // into the builder function when the page loads for the second
    // time. That gives it a chance to build a different version of
    // the form (with additional options, for example). It also
    // does magickal voodoo that keeps validation working properly
    // in those complex multi-page forms. For now, we won't
    // worry about that.
    $form['#multistep'] = TRUE;

    // Setting #redirect to false means that Drupal *won't* attempt
    // to reload a clean copy of the form (by redirecting to the
    // current page, and reloading) when it's submitted. In a multistep
    // form, we want to keep the data around so it can be displayed,
    // or passed on to a subsequent step.
    $form['#redirect'] = FALSE;

    // Add Previous, Next, and/or Finish button(s).
    if ($current_page > 0) { $form['previous'] = array('#type' => 'submit', '#value' => 'Previous', '#weight' => 40); }
    if ($current_page < $pages) { $form['next'] = array('#type' => 'submit', '#value' => 'Next', '#weight' => 45); }
    if ($current_page == $pages) { $form['finish'] = array('#type' => 'submit', '#value' => 'Finish', '#weight' => 50); }

    // Add custom submit function ahead of any others which will short-circuit the FAPI2 workflow.
    $old_submit = (array) $form['#submit'];
    $form['#submit'] = array('wizard_submit' => array(&$form, &$this, (array) $this->fapi_form));
    foreach ($old_submit as $k => $v) {
      $form['#submit'][$k] = $v;
    }

    $this->fapi_form = $form;
    $this->fapi_form['#submit']['wizard_submit'][0] = &$this->fapi_form;
  }

  /**
   * Convert submitted form values from other pages to hidden form elements.
   *
   * @param Array $form
   *   Sliced Forms API form.
   * @param Array $form_values
   *   Submitted form values.
   * @return Array
   *   Hidden form elements.
   */
  function hide_submitted(&$form, &$form_values) {
    // for each $form_value that is not present in $form,
    // add its name and value to an array of hidden form elements

    $elements_only = $this->form_flatten($form); // remove properties and form fieldset elements
    $flattened_elements = $this->form_values_flatten($elements_only); // flatten hierarchy
    $flattened_values = $this->form_values_flatten($form_values);  // flatten hierarchy
    $hidden_elements = $this->array_diff_key($flattened_values, $flattened_elements); // $form_values not present in $form

    if (count($hidden_elements))
    foreach ($hidden_elements as $key => $value) {
      $elements[$key] = array(
        '#type' => 'hidden',
        '#value' => $value,
      );
    }

    return $elements;
  }

  /**
   * Computes the difference of arrays using keys for comparison.
   *
   * @param Array $a1
   *   An array.
   * @param Array $a2 ...
   *   An array.
   * @return Array
   *   An array containing all the values of array1 that have keys that are not present
   *   in any of the other arguments. Note that the associativity is preserved. This function
   *   is like array_diff() except the comparison is done on the keys instead of the values.
   */
  function array_diff_key() {
    $a = func_get_args();

    if (count($a)>1 && is_array($a[0]))
    foreach (array_keys($a[0]) as $k) {
      for ($i=1; $i<count($a); $i++) {
        if (isset($a[$i][$k])) {
          unset($a[0][$k]); // prune the vine
          break;
        }
      }
    }

    return $a[0];
  }

  /**
   * Recursively flatten a $form array.
   *
   * @param Array $elements
   *   A multi-dimensional Forms API form.
   * @return Array
   *   A single-dimensional (flattened) array.
   */
  function form_flatten(&$e, $l=0) {
    static $flattened_form;
    if ($l===0) { $flattened_form = array(); }

    // Recurse through all children.
    foreach (element_children($e) as $k) {
      if (isset($e[$k]) && $e[$k]) {
        if (isset($e[$k]['#tree']) && $e[$k]['#tree']) {
          $flattened_form[$k] = $e[$k];
        } else if ($element = $this->form_flatten($e[$k], $l+1)) {
          if (!isset($element['#type']) || $element['#type'] != 'fieldset') {
            $flattened_form[$k] = $element;
          }
        }
      }
    }

    return $l? $e : $flattened_form;
  }

  /**
   * Recursively flatten a $form_values array.
   *
   * @param Array $a
   *   A multi-dimensional $form['#post'] array.
   * @return Array
   *   A single-dimensional (flattened) array.
   */
  function form_values_flatten(&$a, $l=0) {
    static $breadcrumbs, $elements;
    if ($l===0) { $breadcrumbs = $elements = array(); }

    $children = element_children($a);
    if (is_array($a) && count($children)) {
      foreach ($children as $k) {
        $breadcrumbs[] = $k;
        $this->form_values_flatten($a[$k], 1);
        array_pop($breadcrumbs);
      }
    } else {
       return count($breadcrumbs)?
       $elements[array_pop(array_slice($breadcrumbs, 0, 1)) . (count($breadcrumbs)>1? '['. implode('][', array_slice($breadcrumbs, 1)) .']' : '')] = &$a
       : $a;
    }

    return $elements;
  }
}